/*
 * piano.c - Key piano widget
 *
 * Based on SwamiUiPiano.c from Swami
 * Copyright (C) 1999-2002 Josh Green
 * Modified for specimen by Loki Davison 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA or point your web browser to http://www.gnu.org.
 */

#include <stdio.h>
#include <gtk/gtk.h>
#include <string.h>
#include "piano.h"
#include "patch.h"
#include "specimen.h"
#include "gui.h"
#include "bank-builder.h"

/* Piano keys: C C# D D# E F F# G G# A A# B */

enum
{
     PROP_0,
     PROP_WIDTH_PIXELS,		/* width in pixels */
     PROP_HEIGHT_PIXELS,		/* height in pixels */
     PROP_KEY_COUNT,		/* number of keys in piano */
     PROP_START_OCTAVE,		/* piano start octave (first key) */
     PROP_LOWER_OCTAVE,		/* lower keyboard octave */
     PROP_UPPER_OCTAVE,		/* upper keyboard octave */
};

enum
{
     NOTE_ON,
     NOTE_OFF,
     LAST_SIGNAL
};

/* piano vertical line width to white key width scale */
#define PIANO_VLINE_TO_WHITE_SCALE  (1.0/8.0)

/* piano horizontal line width to piano height scale */
#define PIANO_HLINE_TO_HEIGHT_SCALE (1.0/48.0)

/* piano black key width to white key width */
#define PIANO_BLACK_TO_WHITE_SCALE (5.0/8.0)

/* piano black key height to piano height */
#define PIANO_BLACK_TO_HEIGHT_SCALE (26.0/48.0)

/* piano white key grey edge to piano height scale */
#define PIANO_GREY_TO_HEIGHT_SCALE (2.0/48.0)

/* piano white key indicator width to key scale */
#define PIANO_WHITE_INDICATOR_WIDTH_SCALE (4.0/8.0)

/* piano white key indicator height to piano height scale */
#define PIANO_WHITE_INDICATOR_HEIGHT_SCALE (2.0/48.0)

/* piano black key indicator width to key scale */
#define PIANO_BLACK_INDICATOR_WIDTH_SCALE (3.0/5.0)

/* piano black key indicator height to piano height scale */
#define PIANO_BLACK_INDICATOR_HEIGHT_SCALE (3.0/48.0)

/* piano active black key shorten scale (to look pressed down) */
#define PIANO_BLACK_SHORTEN_SCALE (1.0/26.0)


/* a structure used for each key of a piano */
typedef struct
{
     GnomeCanvasItem *item;      /* canvas item for black keys, white hide rect */
     GnomeCanvasItem *separator; /* white key separator line */
     GnomeCanvasItem *indicator; /* indicator item (if set note is active) */
}
KeyInfo;

static void piano_class_init (PianoClass * klass);
static void piano_set_property (GObject * object, guint property_id,
				const GValue * value, GParamSpec * pspec);
static void piano_get_property (GObject * object, guint property_id,
				GValue * value, GParamSpec * pspec);
static void piano_init (Piano * piano);

static void piano_destroy (GtkObject * object);
static void piano_item_realize (GnomeCanvasItem * item);
static void piano_draw (Piano * piano);
static void piano_draw_noteon (Piano * piano, int note);
static void piano_draw_noteoff (Piano * piano, int note);
static void piano_update_mouse_note (Piano * piano, double x, double y);
static gboolean piano_cb_event (GnomeCanvasItem * item, GdkEvent * event,
				gpointer data);
static gboolean piano_note_on_internal (Piano * piano, int note);
static gboolean piano_note_off_internal (Piano * piano, int note);


static GObjectClass *parent_class = NULL;
static guint piano_signals[LAST_SIGNAL] = { 0 };

GType piano_get_type (void)
{
     static GType obj_type = 0;

     if (!obj_type)
     {
	  static const GTypeInfo obj_info = {
	       sizeof (PianoClass), NULL, NULL,
	       (GClassInitFunc) piano_class_init, NULL, NULL,
	       sizeof (Piano), 0,
	       (GInstanceInitFunc) piano_init,
	  };

	  obj_type = g_type_register_static (GNOME_TYPE_CANVAS_GROUP,
					     "Piano", &obj_info, 0);
     }

     return (obj_type);
}

static void piano_class_init (PianoClass * klass)
{
     GObjectClass *obj_class = G_OBJECT_CLASS (klass);
     GtkObjectClass *gtkobj_class = GTK_OBJECT_CLASS (klass);
     GnomeCanvasItemClass *canvas_item_class =
	  GNOME_CANVAS_ITEM_CLASS (klass);

     parent_class = g_type_class_peek_parent (klass);

     obj_class->set_property = piano_set_property;
     obj_class->get_property = piano_get_property;

     canvas_item_class->realize = piano_item_realize;

     gtkobj_class->destroy = piano_destroy;

     //  klass->note_on = piano_real_note_on;
     //  klass->note_off = piano_real_note_off;

     piano_signals[NOTE_ON] =
	  g_signal_new ("note-on", G_TYPE_FROM_CLASS (klass),
			G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (PianoClass,
							    note_on), NULL,
			NULL, g_cclosure_marshal_VOID__UINT, G_TYPE_NONE, 1,
			G_TYPE_UINT);

     piano_signals[NOTE_OFF] =
	  g_signal_new ("note-off", G_TYPE_FROM_CLASS (klass),
			G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (PianoClass,
							    note_off), NULL,
			NULL, g_cclosure_marshal_VOID__UINT, G_TYPE_NONE, 1,
			G_TYPE_UINT);

     g_object_class_install_property (obj_class, PROP_WIDTH_PIXELS,
				      g_param_spec_int ("width-pixels",
							"Pixel Width",
							"Width in pixels",
							1, G_MAXINT,
							PIANO_DEFAULT_WIDTH,
							G_PARAM_READWRITE));
     g_object_class_install_property (obj_class, PROP_HEIGHT_PIXELS,
				      g_param_spec_int ("height-pixels",
							"Pixel Height",
							"Height in pixels",
							1, G_MAXINT,
							PIANO_DEFAULT_HEIGHT,
							G_PARAM_READWRITE));
     g_object_class_install_property (obj_class, PROP_KEY_COUNT,
				      g_param_spec_int ("key-count",
							"Key Count",
							"Size of piano in keys",
							1, 128, 128,
							G_PARAM_READWRITE |
							G_PARAM_CONSTRUCT_ONLY));
     g_object_class_install_property (obj_class, PROP_START_OCTAVE,
				      g_param_spec_int ("start-octave",
							"Start Octave",
							"Piano start octave (0 = MIDI note 0",
							0, 10, 0,
							G_PARAM_READWRITE));
     g_object_class_install_property (obj_class, PROP_LOWER_OCTAVE,
				      g_param_spec_int ("lower-octave",
							"Lower Octave",
							"Lower keyboard start octave",
							0, 10,
							PIANO_DEFAULT_LOWER_OCTAVE,
							G_PARAM_READWRITE));
     g_object_class_install_property (obj_class, PROP_UPPER_OCTAVE,
				      g_param_spec_int ("upper-octave",
							"Upper Octave",
							"Upper keyboard start octave",
							0, 10,
							PIANO_DEFAULT_UPPER_OCTAVE,
							G_PARAM_READWRITE));

}

/*Create a new piano widget */
GtkWidget *piano_new (void)
{
     return GTK_WIDGET (g_object_new (TYPE_PIANO, NULL));
}


static void
piano_set_property (GObject * object, guint property_id,
		    const GValue * value, GParamSpec * pspec)
{
     Piano *piano = PIANO (object);

     switch (property_id)
     {
     case PROP_WIDTH_PIXELS:
	  piano->width = g_value_get_int (value);
	  piano->up2date = FALSE;
	  piano_draw (piano);
	  break;
     case PROP_HEIGHT_PIXELS:
	  piano->height = g_value_get_int (value);
	  piano->up2date = FALSE;
	  piano_draw (piano);
	  break;
     case PROP_KEY_COUNT:
	  piano->key_count = g_value_get_int (value);
	  break;
     case PROP_START_OCTAVE:
	  piano->start_note = g_value_get_int (value) * 12;
	  break;
     case PROP_LOWER_OCTAVE:
	  piano->lower_octave = g_value_get_int (value);
	  break;
     case PROP_UPPER_OCTAVE:
	  piano->upper_octave = g_value_get_int (value);
	  break;
     default:
	  G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	  break;
     }
}

static void
piano_get_property (GObject * object, guint property_id,
		    GValue * value, GParamSpec * pspec)
{
     Piano *piano = PIANO (object);

     switch (property_id)
     {
     case PROP_WIDTH_PIXELS:
	  g_value_set_int (value, piano->width);
	  break;
     case PROP_HEIGHT_PIXELS:
	  g_value_set_int (value, piano->height);
	  break;
     case PROP_KEY_COUNT:
	  g_value_set_int (value, piano->key_count);
	  break;
     case PROP_START_OCTAVE:
	  g_value_set_int (value, piano->start_note / 12);
	  break;
     case PROP_LOWER_OCTAVE:
	  g_value_set_int (value, piano->lower_octave);
	  break;
     case PROP_UPPER_OCTAVE:
	  g_value_set_int (value, piano->upper_octave);
	  break;
     default:
	  G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	  break;
     }
}

static void piano_init (Piano * piano)
{
     int i;

     piano->key_count = 128;	/* FIXME: Shouldn't it get set on construction? */

     piano->width = PIANO_DEFAULT_WIDTH;
     piano->height = PIANO_DEFAULT_HEIGHT;
     piano->key_info = g_new0 (KeyInfo, piano->key_count);

     piano->start_note = 0;
     piano->lower_octave = PIANO_DEFAULT_LOWER_OCTAVE;
     piano->upper_octave = PIANO_DEFAULT_UPPER_OCTAVE;
     piano->mouse_note = 128;	/* disabled mouse key (> 127) */

     /* get the white key count for the number of keys in the piano */
     piano->white_count = piano->key_count / 12 * 7;	/* 7 whites per oct */
     i = piano->key_count % 12;	/* calculate for octave remainder (if any) */
     if (i < 6)
	  piano->white_count += (i + 1) / 2;
     else
	  piano->white_count += (i | 0x1) / 2 + 1;


     g_signal_connect (G_OBJECT (piano),
		       "event", G_CALLBACK (piano_cb_event), piano);
}

static void piano_destroy (GtkObject * object)
{
     Piano *piano = PIANO (object);

     g_free (piano->key_info);
     piano->key_info = NULL;

     if (GTK_OBJECT_CLASS (parent_class)->destroy)
	  (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void piano_item_realize (GnomeCanvasItem * item)
{
     Piano *piano = PIANO (item);

     if (GNOME_CANVAS_ITEM_CLASS (parent_class)->realize)
	  (*GNOME_CANVAS_ITEM_CLASS (parent_class)->realize) (item);

     piano_draw (piano);
}

static void piano_draw (Piano * piano)
{
     GnomeCanvasGroup *group = GNOME_CANVAS_GROUP (piano);
     GnomeCanvasItem *item;
     GnomeCanvasPoints *points;
     KeyInfo *key_info;
     double x;
     int i, mod, note_ndx;

     /* return if not added to a canvas yet */
     if (!GNOME_CANVAS_ITEM (piano)->canvas)
	  return;

     if (!piano->up2date)
     {
	  /* convert pixel width and height to world values */
	  gnome_canvas_c2w (GNOME_CANVAS_ITEM (piano)->canvas,
			    piano->width - 1, piano->height - 1,
			    &piano->world_width, &piano->world_height);

	  piano->whitekey_width = piano->world_width / piano->white_count;
	  piano->vline_width =
	       piano->whitekey_width * PIANO_VLINE_TO_WHITE_SCALE;
	  piano->hline_width = piano->height * PIANO_HLINE_TO_HEIGHT_SCALE;

	  /* black key dimensions */
	  piano->black_half = piano->whitekey_width
	       * PIANO_BLACK_TO_WHITE_SCALE / 2.0;
	  piano->black_height =
	       piano->world_height * PIANO_BLACK_TO_HEIGHT_SCALE;

	  /* top of grey edge */
	  piano->grey_top = piano->world_height
	       - piano->world_height * PIANO_GREY_TO_HEIGHT_SCALE;

	  piano->up2date = TRUE;
     }

     if (!piano->white_bg)	/* one time canvas item creation */
     {
	  piano->white_bg =	/* white key background */
	       gnome_canvas_item_new (group, GNOME_TYPE_CANVAS_RECT,
				      "x1", 0.0,
				      "y1", 0.0,
				      "fill-color", "white",
				      "outline-color", NULL, NULL);

	  /* bottom grey edge of white keys */
	  piano->grey_edge =
	       gnome_canvas_item_new (group, GNOME_TYPE_CANVAS_RECT,
				      "x1", 0.0,
				      "fill-color", "grey",
				      "outline-color", NULL, NULL);

	  /* create white key divider lines and black keys */
	  key_info = (KeyInfo *) (piano->key_info);
	  for (i = 1, mod = 1, note_ndx = 0; i < piano->white_count;
	       i++, mod++)
	  {
	       key_info[note_ndx].separator =	/* white key divider line */
		    gnome_canvas_item_new (group, GNOME_TYPE_CANVAS_LINE,
					   "fill-color", "black", NULL);

	       if (mod > 7)
		    mod = 1;	/* white key octave modulo */

	       /* black keys */
	       if (mod != 3 && mod != 7)
	       {			/* create black key */
		    item =
			 gnome_canvas_item_new (group, GNOME_TYPE_CANVAS_RECT,
						"y1", 0.0, "fill-color",
						"black", NULL);
		    key_info[note_ndx + 1].item = item;	/* store item in key_info */
		    note_ndx += 2;	/* index of next white key */
	       }
	       else
		    note_ndx++;	/* index of next white key */
	  }

	  /* piano outline */
	  piano->outline =
	       gnome_canvas_item_new (group, GNOME_TYPE_CANVAS_RECT,
				      "x1", 0.0,
				      "y1", 0.0,
				      "fill-color", NULL,
				      "outline-color", "black", NULL);
     }

     gnome_canvas_item_set (piano->white_bg,	/* set white background size */
			    "x2", piano->world_width,
			    "y2", piano->grey_top,
			    "width-units", piano->hline_width, NULL);

     gnome_canvas_item_set (piano->grey_edge,	/* set white key grey edge size */
			    "y1", piano->grey_top,
			    "x2", piano->world_width,
			    "y2", piano->world_height,
			    "width-units", piano->hline_width, NULL);

     points = gnome_canvas_points_new (2);

     key_info = (KeyInfo *) (piano->key_info);
     for (i = 1, mod = 1, note_ndx = 0; i < piano->white_count; i++, mod++)
     {
	  x = i * piano->whitekey_width;	/* position of white key separator */

	  points->coords[0] = x;
	  points->coords[1] = 0.0;
	  points->coords[2] = x;
	  points->coords[3] = piano->world_height;

	  /* white key divider line */
	  gnome_canvas_item_set (key_info[note_ndx].separator,
				 "points", points,
				 "width-units", piano->vline_width, NULL);

	  if (mod > 7)
	       mod = 1;		/* white key octave modulo */

	  /* black keys */
	  if (mod != 3 && mod != 7)
	  {			/* create black key */
	       gnome_canvas_item_set (key_info[note_ndx + 1].item,
				      "x1", x - piano->black_half,
				      "x2", x + piano->black_half,
				      "y2", piano->black_height, NULL);
	       note_ndx += 2;	/* index of next white key */
	  }
	  else
	       note_ndx++;		/* index of next white key */
     }

     gnome_canvas_points_free (points);

     gnome_canvas_item_set (piano->outline,	/* set outline size */
			    "x2", piano->world_width,
			    "y2", piano->world_height,
			    "width-units", piano->hline_width, NULL);
}

static void piano_draw_noteon (Piano * piano, int note)
{
     GnomeCanvasGroup *group = GNOME_CANVAS_GROUP (piano);

     KeyInfo *key_info;
     double white_x, w, h, x;
     int note_ofs, white_index, mod, cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     /* get key info structure for note */
     note_ofs = note - piano->start_note;
     key_info = &((KeyInfo *) (piano->key_info))[note_ofs];

     /* calculate white key x coordinate */
     white_index = note_ofs / 12 * 7;	/* 7 white keys per octave */
     mod = note_ofs % 12;
     if (mod <= 4)
	  white_index += mod / 2;
     else
	  white_index += (mod - 1) / 2 + 1;

     white_x = white_index * piano->whitekey_width;

     if (!key_info->item)	/* white key? */
     {
	  key_info->item =	/* store white hide rectangle to unused item field */
	       gnome_canvas_item_new (group, GNOME_TYPE_CANVAS_RECT,
				      "x1", white_x,
				      "y1", piano->grey_top,
				      "x2", white_x + piano->whitekey_width,
				      "y2", piano->world_height,
				      "fill-color", "green",
				      "outline-color", NULL, NULL);

	  /* put hide rectangle just above grey key tip */
	  gnome_canvas_item_lower_to_bottom (key_info->item);
	  gnome_canvas_item_raise (key_info->item, 2);

	  x = white_x + piano->whitekey_width / 2.0;	/* center of white key */

	  /* calculate size of white key indicator */
	  w = piano->whitekey_width * PIANO_WHITE_INDICATOR_WIDTH_SCALE / 2;
	  h = piano->world_height / 2.0;

	  /* create the indicator if the current note is the root note and move it into place to ease
	   * translation later. */
	  if (note == patch_get_note (cp))
	  {
	       key_info->indicator =
		    gnome_canvas_item_new (group, GNOME_TYPE_CANVAS_RECT,
					   "x1", x - w,
					   "y1", h,
					   "x2", x + w + 1,
					   "y2", h + 10,
					   "fill-color", "dark green",
					   "outline-color", NULL, NULL);
	       gnome_canvas_item_move (key_info->indicator, 0.0, 3.0);
	  }
     }
     else			/* black key */
     {				/* shorten black key to make it look pressed down */
	  gnome_canvas_item_set (GNOME_CANVAS_ITEM (key_info->item),
				 "y2", piano->black_height
				 - piano->black_height
				 * PIANO_BLACK_SHORTEN_SCALE, "fill-color",
				 "dark green", NULL);

	  /* x position of center of black key */
	  x = white_x + piano->whitekey_width;

	  /* calculate size of black key indicator */
	  w = piano->black_half * PIANO_BLACK_INDICATOR_WIDTH_SCALE;
	  h = piano->black_height * (PIANO_BLACK_INDICATOR_HEIGHT_SCALE /
				     2.0);

	  if (note == patch_get_note (cp))
	  {
	       /* create the indicator and move it into place to ease
		* translation later */
	       key_info->indicator =
		    gnome_canvas_item_new (group, GNOME_TYPE_CANVAS_RECT,
					   "x1", x - w,
					   "y1", -h,
					   "x2", x + w,
					   "y2", h * 2,
					   "fill-color", "yellow",
					   "outline-color", NULL, NULL);
	       gnome_canvas_item_move (key_info->indicator, 0.0, 5.0);
	  }
     }
}

/* Draw what note range is currently selected. */
void piano_draw_range (Piano * piano)
{
     int i = 0;
     int cp;

     /* turn of all the keys then draw the new region */
     for (i = 0; i < piano->key_count; i++)
     {
	  piano_note_off_internal (piano, i);
     }

     if ((cp = get_current_patch ( )) < 0)
	  return;
     if (patch_get_range (cp))
     {

	  for (i = patch_get_lower_note (cp); i <= patch_get_upper_note (cp);
	       i++)
	  {
	       piano_note_on_internal (piano, i);
	  }

     }
     else
     {
	  i = patch_get_note (cp);
	  piano_note_on_internal (piano, i);
     }
}

static void piano_draw_noteoff (Piano * piano, int note)
{
     KeyInfo *key_info;
     int note_ofs, mod;
     gboolean black;

     /* get key info structure for note */
     note_ofs = note - piano->start_note;
     key_info = &((KeyInfo *) (piano->key_info))[note_ofs];

     if (key_info->indicator)
     {
	  /* destroy indicator */
	  gtk_object_destroy (GTK_OBJECT (key_info->indicator));
	  key_info->indicator = NULL;
     }

     /* determine if key is black or white */
     mod = note % 12;
     if (mod <= 4)
	  black = mod & 1;
     else
	  black = !(mod & 1);

     if (black)
     {				/* reset black key back to normal length */
	  gnome_canvas_item_set (GNOME_CANVAS_ITEM (key_info->item),
				 "y2", piano->black_height, "fill-color",
				 "black", NULL);
     }
     else
     {				/* destroy hide rectangle */
	  if (key_info->item)
	  {
	       gtk_object_destroy (GTK_OBJECT (key_info->item));
	       key_info->item = NULL;
	  }
     }
}

/* Called when dragging the mouse around with a button held down on the piano */
static void piano_update_mouse_note (Piano * piano, double x, double y)
{
     KeyInfo *key_info;
     gboolean white = TRUE;
     int white_keynum, mod;
     int note;

     /* start and end keys assumed to be white */
     if (x < 0.0)
	  note = piano->start_note;
     else if (x > piano->world_width)
	  note = piano->start_note * piano->key_count;
     else
     {				/* calculate note number from white key index from x position */
	  white_keynum = x / piano->whitekey_width;
	  note = white_keynum / 7 * 12;
	  mod = white_keynum % 7;
	  if (mod < 3)
	       note += mod * 2;
	  else
	       note += mod * 2 - 1;

	  /* possible black key or off the top or bottom edge of the piano */
	  if (y <= piano->black_height || y > piano->world_height)
	  {			/* offset, in canvas world units, within white key */
	       double key_ofs = x - (white_keynum * piano->whitekey_width);

	       /* check for black key to the left of white key */
	       if (note != 0 && mod != 0 && mod != 3
		   && key_ofs < piano->black_half)
	       {
		    white = FALSE;
		    note--;
	       }			/* black key to the right of white key? */
	       else if (note < (piano->key_count - 1) && mod != 2 && mod != 6
			&& key_ofs >
			(piano->whitekey_width - piano->black_half))
	       {
		    white = FALSE;
		    note++;
	       }
	  }
     }

     key_info = &((KeyInfo *) (piano->key_info))[note];

     note += piano->start_note;

     if (note > 127)
     {
	  debug ("off end of piano changing to 127 \n");
	  note = 127;
     }
     if (note < 0)
     {
	  debug ("off end of piano changing to 0 \n");
	  note = 0;
     }

     if (note != piano->mouse_note)	/* note changed? */
     {
	  piano_note_off (piano, piano->mouse_note);	/* note off */
	  piano->mouse_note = note;
	  piano_note_on (piano, note, 1);	/* new note on */
     }
}


/* Converts a the list of uri we are droped into gslist. 
 * Makes them the right format for filenames etc too.
 * This should probably go into the utils file in the next merge. */

GSList *uri_list_to_gslist (char *uri_list)
{
     GSList *list = NULL;

     /* process each entry in the list */
     while (*uri_list)
     {
	  char *linebreak;
	  char *uri;
	  int length;

	  linebreak = strchr (uri_list, 13);

	  /* return if it's the end */
	  if (!linebreak || linebreak[1] != 10)
	  {
	       return list;
	  }

	  length = linebreak - uri_list;

	  if (length && uri_list[0] != '#')
	  {
	       uri = g_malloc (sizeof (char) * (length + 1));
	       strncpy (uri, uri_list, length);
	       uri[length] = 0;
	       /* get rid of the file:/// crap so soundfile likes it. */
	       list =
		    g_slist_append (list,
				    g_filename_from_uri (uri, NULL, NULL));
	  }

	  uri_list = linebreak + 2;
     }
     return list;
}

/* handle drag and drop events */

void piano_dnd (GtkWidget * widget, GdkDragContext * context,
		gint x, gint y, GtkSelectionData * selection_data,
		guint info, guint time, gpointer data)
{
     debug ("we got a drop! \n");
     GSList *filelist = NULL;
     int note;
     Piano *piano = PIANO (data);

     note = piano_pos_to_note (piano, x, y, NULL);
     if (!selection_data->data)
     {
	  return;
     }

     /* get filename then create new patch, selection data is actually the filenames all as text,
      * seperated by new lines so if we want multifile support we have to process that... */
     /* the file selector gives us a uri, so we have to trim the file:/// stuff off it. */
     filelist = uri_list_to_gslist (selection_data->data);
     bb_new_patch (note, filelist);
     //prehaps we should free the filelist here? 

}

static gboolean
piano_cb_event (GnomeCanvasItem * item, GdkEvent * event, gpointer data)
{
     Piano *piano = PIANO (data);
     GdkEventButton *bevent;
     GdkEventMotion *mevent;
     int note;

     switch (event->type)
     {
     case GDK_BUTTON_PRESS:
	  bevent = (GdkEventButton *) event;
	  //if (bevent->button != 1) break; // handle all buttons
	  note = piano_pos_to_note (piano, bevent->x, bevent->y, NULL);
	  if (note < 0)
	       break;		/* click not on a note? */
	  piano->mouse_note = note;
	  piano_note_on (piano, note, bevent->button);

	  gnome_canvas_item_grab (item, GDK_POINTER_MOTION_MASK
				  | GDK_BUTTON_RELEASE_MASK, NULL,
				  bevent->time);
	  return (TRUE);
     case GDK_BUTTON_RELEASE:
	  if (piano->mouse_note > 127)
	       break;		/* no mouse selected note? */
	  note = piano->mouse_note;
	  piano->mouse_note = 128;	/* no more selected mouse note */
	  piano_note_off (piano, note);
	  piano_draw_range (piano);	/*draw current note range */
	  gnome_canvas_item_ungrab (item, event->button.time);
	  break;
     case GDK_MOTION_NOTIFY:
	  if (piano->mouse_note > 127)
	       break;		/* no mouse selected note? */
	  mevent = (GdkEventMotion *) event;
	  piano_update_mouse_note (piano, mevent->x, mevent->y);
	  break;
     default:
	  break;
     }

     return (FALSE);
}

/**
 * piano_note_on:
 * @piano: Piano object
 * Piano note on. Turns on note visually on piano as well set the root key
 * Also updates the gui spinbuttons with the new note info. 
 */

void piano_note_on (Piano * piano, int note, int button)
{
     int val, cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;
     if (button == 1)
     {
	  if ((val = patch_set_note (cp, note)) < 0)
	  {
	       fprintf (stderr, "Failed to set patch %d to note %d (%s)\n",
			cp, note, patch_strerror (val));
	  }
	  set_note_spin (note);
	  printf ("Patch %d now listening on note %d\n", cp, note);
     }
     else
     {
	  patch_set_range (cp, 1);
	  set_range_active ( );
	  debug ("Patch %d now listening to a range of notes\n", cp);
	  if (note < patch_get_note (cp))
	  {
	       if ((val = patch_set_lower_note (cp, note)) < 0)
	       {
		    debug ("Unable to set patch %d lower note to %d (%s)\n",
			   cp, note, patch_strerror (val));
		    return;
	       }
	       set_range_low_spin (note);
	       debug ("Patch %d lower note set to %d\n", cp, note);
	  }
	  else if (note > patch_get_note (cp))
	  {
	       if ((val = patch_set_upper_note (cp, note)) < 0)
	       {
		    debug ("Unable to set patch %d upper note to %d (%s)\n",
			   cp, note, patch_strerror (val));
		    return;
	       }
	       set_range_up_spin (note);
	       debug ("Patch %d upper note set to %d\n", cp, note);
	  }
     }
     piano_note_on_internal (piano, note);
}

/* internal function for visually displaying a note on */
static gboolean piano_note_on_internal (Piano * piano, int note)
{
     KeyInfo *key_info;
     int start_note;

     g_return_val_if_fail (IS_PIANO (piano), FALSE);
     g_return_val_if_fail (note >= 0 && note <= 127, FALSE);
     start_note = piano->start_note;
     if (note < start_note || note >= start_note + piano->key_count)
	  return (FALSE);

     key_info = &((KeyInfo *) (piano->key_info))[note - start_note];
     if (key_info->indicator)
	  return (FALSE);		/* note already on? */

     piano_draw_noteon (piano, note);
     return (TRUE);
}

/**
 * piano_note_off:
 * @piano: Piano object
 * @note: MIDI note number to turn off
 * Piano note off. Turns off note visually on piano.
 */
void piano_note_off (Piano * piano, int note)
{
     if (piano_note_off_internal (piano, note))
     {
     }
}

/* updates drawing of piano */
static gboolean piano_note_off_internal (Piano * piano, int note)
{
     KeyInfo *key_info;
     int start_note;

     g_return_val_if_fail (IS_PIANO (piano), FALSE);
     g_return_val_if_fail (note >= 0 && note <= 127, FALSE);
     start_note = piano->start_note;
     if (note < start_note || note >= start_note + piano->key_count)
	  return (FALSE);

     key_info = &((KeyInfo *) (piano->key_info))[note - start_note];

     piano_draw_noteoff (piano, note);
     return (TRUE);
}

/**
 * piano_pos_to_note:
 * @piano: Piano object
 * @x: X coordinate in canvas world units
 * @y: Y coordinate in canvas world units
 * @isblack: Location to store a boolean if key is black/white, or %NULL
 */
int
piano_pos_to_note (Piano * piano, double x, double y, gboolean * isblack)
{
     gboolean white = TRUE;
     double key_ofs;
     int white_keynum, mod;
     int note;

     g_return_val_if_fail (IS_PIANO (piano), -1);

     /* within piano bounds? */
     if (x < 0.0 || x > piano->world_width || y < 0.0
	 || y > piano->world_height)
	  return (-1);

     /* calculate note number from white key index from x position */
     white_keynum = x / piano->whitekey_width;
     note = white_keynum / 7 * 12;
     mod = white_keynum % 7;
     if (mod < 3)
	  note += mod * 2;
     else
	  note += mod * 2 - 1;

     if (y <= piano->black_height)	/* possible black key? */
     {				/* offset, in canvas world units, within white key */
	  key_ofs = x - (white_keynum * piano->whitekey_width);

	  /* check for black key to the left of white key */
	  if (note != 0 && mod != 0 && mod != 3
	      && key_ofs < piano->black_half)
	  {
	       white = FALSE;
	       note--;
	  }			/* black key to the right of white key? */
	  else if (note < (piano->key_count - 1) && mod != 2 && mod != 6
		   && key_ofs > (piano->whitekey_width - piano->black_half))
	  {
	       white = FALSE;
	       note++;
	  }
     }
     if (piano->start_note + note >= 0 && piano->start_note + note <= 127)
     {
	  if (isblack)
	       *isblack = !white;
	  return (piano->start_note + note);
     }
     else
     {
	  if (piano->start_note + note < 0)
	  {
	       /* gone of the end of the piano somehow, change it to zero */
	       return 0;
	  }
	  else
	  {
	       /* or the other end, change to 127 */
	       return 127;
	  }
     }
}

/**
 * piano_note_to_pos (Piano *piano, int note, gboolean isblack)
 * @piano: Piano object
 * @isblack: Pointer to store a gboolean which is set if key is black, or %NULL
 *
 * Find the canvas X coordinate for a given MIDI note.
 *
 * Returns: X canvas coordinate to center of @note.
 */
double piano_note_to_pos (Piano * piano, int note, gboolean * isblack)
{
     double pos;
     int white_index, mod;
     gboolean black;

     /* calculate number of white keys to note */
     white_index = note / 12 * 7;
     mod = note % 12;
     if (mod <= 4)
	  white_index += mod / 2;
     else
	  white_index += (mod - 1) / 2 + 1;

     /* determine if key is black or not */
     if (mod <= 4)
	  black = mod & 1;
     else
	  black = !(mod & 1);

     pos = white_index * piano->whitekey_width;
     if (black)
	  pos += piano->whitekey_width;
     else
	  pos += piano->whitekey_width / 2;

     if (isblack)
	  *isblack = black;
     return (pos);
}
